/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>

#include "SDL.h"
#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_sysvideo.h"
#include "SDL_pixels_c.h"
#include "SDL_events_c.h"

#include "SDL_os2video.h"

#include "SDL_vkeys.h" // Some more VK_ codes

// Configuration defines:

// Define the following if you want to allow every resolution if
// one is requesting for a windowed resolution.
// Not recommended, as we can switch only to some given
// fullscreen resolutions, so ToggleFullscreen() would not
// work for some modes.
// If undefined, only those resolutions will be allowed in
// windowed mode that are available as fullscreen mode too!
//#define ALLOW_EVERY_MODE_IN_WINDOWED



/* The translation table from a VK keysym to a SDL keysym */
static SDLKey VK_keymap[SDLK_LAST];
static SDL_keysym *TranslateKey(int vkey, int chcode, int scancode, SDL_keysym *keysym, int iPressed);

#ifdef DEBUG_BUILD
/////////////////////////////////////////////////////////////////////
//
// PrintModeInfo
//
// This is a debug function to print
// a given mode info
//
/////////////////////////////////////////////////////////////////////
static void PrintModeInfo(PGDDMODEINFO OneModeInfo)
{
  if (!OneModeInfo) return;

  printf("Mode ID: 0x%x", OneModeInfo->ulModeId);
  printf(" %dx%d/%d (%d Hz)",
	 OneModeInfo->ulHorizResolution,
	 OneModeInfo->ulVertResolution,
	 OneModeInfo->ulBpp,
	 OneModeInfo->ulRefreshRate);
  printf(" Colors: %d FourCC: %c%c%c%c\n",
	 OneModeInfo->cColors,
	 OneModeInfo->fccColorEncoding & 0xff,
	 (OneModeInfo->fccColorEncoding & 0xff00) >> 8,
         (OneModeInfo->fccColorEncoding & 0xff0000) >> 16,
	 (OneModeInfo->fccColorEncoding & 0xff000000) >> 24);
  fflush(stdout);
}
#endif

/////////////////////////////////////////////////////////////////////
//
// InspectDIVE
//
// Asks DIVE about its capabilities,
// like usable (known) input buffer
// fourCCs and others...
//
/////////////////////////////////////////////////////////////////////
static int InspectDIVE(SDL_VideoDevice *pVideo)
{
  APIRET rc;
  int i;

#ifdef DEBUG_BUILD
  printf("InspectDIVE()\n");
  fflush(stdout);
#endif

  // Our results are stored in our global instance.
  // If for some reason we'd be called twice or more,
  // we should check if the stuffs are already allocated
  // or not, to prevent memory leak!
  if (pVideo->hidden->pDiveCaps)
  {
    free(pVideo->hidden->pDiveCaps);
    pVideo->hidden->pDiveCaps = NULL;
  }
  if (pVideo->hidden->pfccFormats)
  {
    free(pVideo->hidden->pfccFormats);
    pVideo->hidden->pfccFormats = NULL;
  }

  // Allocate needed memory
  pVideo->hidden->pDiveCaps = (PDIVE_CAPS) malloc(sizeof(DIVE_CAPS));
  if (!(pVideo->hidden->pDiveCaps)) return FALSE;
  pVideo->hidden->pfccFormats = (ULONG *)malloc(sizeof(ULONG)*100);
  if (!(pVideo->hidden->pfccFormats))
  {
    free(pVideo->hidden->pDiveCaps);
    return FALSE;
  }
  // Fill some vital information
  pVideo->hidden->pDiveCaps->ulStructLen = sizeof(DIVE_CAPS);
  pVideo->hidden->pDiveCaps->ulFormatLength = sizeof(ULONG)*100;
  pVideo->hidden->pDiveCaps->pFormatData = pVideo->hidden->pfccFormats;
  // Query DIVE capabilities
  rc=DiveQueryCaps(pVideo->hidden->pDiveCaps, DIVE_BUFFER_SCREEN);
  if (rc)
  { // No DIVE on this computer???
#ifdef DEBUG_BUILD
    printf("  DiveQueryCaps(), rc=%d\n", rc);
    fflush(stdout);
#endif

    free(pVideo->hidden->pfccFormats);
    free(pVideo->hidden->pDiveCaps);
    return FALSE;
  }

#ifdef DEBUG_BUILD
  {
    char *fcc = (char *) &(pVideo->hidden->pDiveCaps->fccColorEncoding);

    printf(" DIVE Information:\n");
    printf("   Current screen mode: %d x %d, %d BPP, colour encoding %c%c%c%c\n",
	   pVideo->hidden->pDiveCaps->ulHorizontalResolution,
	   pVideo->hidden->pDiveCaps->ulVerticalResolution,
	   pVideo->hidden->pDiveCaps->ulDepth,
	   *fcc,
	   *(fcc+1),
	   *(fcc+2),
	   *(fcc+3));
    printf("   Number of input formats : %d\n",pVideo->hidden->pDiveCaps->ulInputFormats);
    printf("   Number of output formats: %d\n",pVideo->hidden->pDiveCaps->ulOutputFormats);
    printf("   Input formats:  ");

    fcc=(char *)(pVideo->hidden->pfccFormats);
    for (i=0; i<pVideo->hidden->pDiveCaps->ulInputFormats; i++)
    {
      printf(" %c%c%c%c",
	     *(fcc),
	     *(fcc+1),
	     *(fcc+2),
	     *(fcc+3));
      fcc+=4;
    }
    printf("\n");
    printf("   Output formats: ");
    for (i=0; i<pVideo->hidden->pDiveCaps->ulOutputFormats; i++)
    {
      printf(" %c%c%c%c",
	     *(fcc),
	     *(fcc+1),
	     *(fcc+2),
	     *(fcc+3));
      fcc+=4;
    }
    printf("\n");
    fflush(stdout);
  }

#endif
  return 1;
}

/////////////////////////////////////////////////////////////////////
//
// SetAccessableWindowPos
//
// Same as WinSetWindowPos(), but takes care for the window to be
// always on the screen, the titlebar will be accessable everytime.
//
/////////////////////////////////////////////////////////////////////
static BOOL SetAccessableWindowPos(HWND hwnd, HWND hwndInsertBehind,
				   LONG x, LONG y,
				   LONG cx, LONG cy,
                                   ULONG fl)
{
  SWP swpDesktop, swp;
  // Get desktop area
  WinQueryWindowPos(HWND_DESKTOP, &swpDesktop);

  if ((fl & SWP_MOVE) && (fl & SWP_SIZE))
  {
    // If both moving and sizing, then change size and pos now!!
    if (x>swpDesktop.cx)
      x = 0;
    if (x+cx<0)
      x = 0;
    if (y<0)
      y = 0;
    if (y+cy>swpDesktop.cy)
      y = swpDesktop.cy - cy;
    return WinSetWindowPos(hwnd, hwndInsertBehind, x, y, cx, cy, fl);
  } else
  if (fl & SWP_MOVE)
  {
    // Just moving
    WinQueryWindowPos(hwnd, &swp);
    if (x>swpDesktop.cx)
      x = 0;
    if (x+swp.cx<0)
      x = 0;
    if (y<0)
      y = 0;
    if (y+swp.cy>swpDesktop.cy)
      y = swpDesktop.cy - swp.cy;
    return WinSetWindowPos(hwnd, hwndInsertBehind, x, y, cx, cy, fl);
  } else
  if (fl & SWP_SIZE)
  {
    // Just moving
    WinQueryWindowPos(hwnd, &swp);
    if (swp.x>swpDesktop.cx)
      x = 0;
    if (swp.x+cx<0)
      x = 0;
    if (swp.y<0)
      y = 0;
    if (swp.y+cy>swpDesktop.cy)
      y = swpDesktop.cy - cy;
    return WinSetWindowPos(hwnd, hwndInsertBehind, x, y, cx, cy, fl);
  } else
  return WinSetWindowPos(hwnd, hwndInsertBehind, x, y, cx, cy, fl);
}

/////////////////////////////////////////////////////////////////////
//
// ContainsFCC
//
// Taking an array of FourCCs and the number of array elements,
// it checks if a given FourCC is in the buffer, or not.
//
/////////////////////////////////////////////////////////////////////
static int ContainsFCC(ULONG *pfccList, int iListSize, ULONG fcc)
{
  int i;
  for (i=0; i<iListSize; i++)
  {
    if (*pfccList==fcc)
      return 1;
    pfccList++;
  }
  return 0;
}

/////////////////////////////////////////////////////////////////////
//
// NearestAcceptableDiveInputFCC
//
// Returns a FourCC, which is the closest match to the given one,
// and exists in the list of supported DIVE input formats.
//
/////////////////////////////////////////////////////////////////////
static ULONG NearestAcceptableDiveInputFCC(SDL_VideoDevice *pVideo, ULONG fccFourCC)
{
  switch (fccFourCC)
  {
    case FOURCC_R565:
      /* 5-6-5 RGB*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      return fccFourCC;

    case FOURCC_R555:
      /* 5-5-5 RGB*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      return fccFourCC;
    case FOURCC_R666:
      /* 6-6-6 RGB*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      return fccFourCC;
    case FOURCC_R664:
      /* 6-6-4 RGB*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      return fccFourCC;
    case FOURCC_RGB3:
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      return fccFourCC;
    case FOURCC_BGR3:
        /* 8-8-8 RGB*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      return fccFourCC;

    case FOURCC_RGB4:
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      return fccFourCC;

    case FOURCC_BGR4:
      /* 8-8-8 BGR*/
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR4)) return FOURCC_BGR4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB4)) return FOURCC_RGB4;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_BGR3)) return FOURCC_BGR3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_RGB3)) return FOURCC_RGB3;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R565)) return FOURCC_R565;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R555)) return FOURCC_R555;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R664)) return FOURCC_R664;
      if (ContainsFCC(pVideo->hidden->pfccFormats,
		      pVideo->hidden->pDiveCaps->ulInputFormats,
		      FOURCC_R666)) return FOURCC_R666;
      return fccFourCC;
    default:
      return fccFourCC;
  }
}

/////////////////////////////////////////////////////////////////////
//
// GetBitMask
//
// Returns the R G and B bitmask plus the BitsPerPixel value for
// a given FourCC.
//
/////////////////////////////////////////////////////////////////////
static void GetBitMask(ULONG fccFourCC, unsigned int *puiRmask, unsigned int *puiGmask, unsigned int *puiBmask, int *piBPP)
{
  switch (fccFourCC)
  {
    case FOURCC_R565:
        /* 5-6-5 RGB*/
        *puiRmask = 0x0000f800;
        *puiGmask = 0x000007e0;
	*puiBmask = 0x0000001f;
        *piBPP = 16;
        break;
    case FOURCC_R555:
        /* 5-5-5 RGB*/
        *puiRmask = 0x00007c00;
        *puiGmask = 0x000003e0;
	*puiBmask = 0x0000001f;
        *piBPP = 15;
        break;
    case FOURCC_R666:
        /* 6-6-6 RGB*/
        *puiRmask = 0x0003f000;
        *puiGmask = 0x00000fc0;
	*puiBmask = 0x0000003f;
        *piBPP = 18;
        break;
    case FOURCC_R664:
        /* 6-6-4 RGB*/
        *puiRmask = 0x0000fc00;
        *puiGmask = 0x000003f0;
	*puiBmask = 0x0000000f;
        *piBPP = 16;
        break;
    case FOURCC_RGB3:
        /* 8-8-8 RGB*/
        *puiRmask = 0x00ff0000;
        *puiGmask = 0x0000ff00;
	*puiBmask = 0x000000ff;
        *piBPP = 24;
    case FOURCC_RGB4:
        /* 8-8-8 RGB*/
        *puiRmask = 0x00ff0000;
        *puiGmask = 0x0000ff00;
	*puiBmask = 0x000000ff;
        *piBPP = 32;
        break;
    case FOURCC_BGR3:
        /* 8-8-8 BGR*/
        *puiRmask = 0x000000ff;
        *puiGmask = 0x0000ff00;
	*puiBmask = 0x00ff0000;
        *piBPP = 24;
        break;
    case FOURCC_BGR4:
        /* 8-8-8 BGR*/
        *puiRmask = 0x000000ff;
        *puiGmask = 0x0000ff00;
	*puiBmask = 0x00ff0000;
        *piBPP = 32;
        break;
    default:
        // We treat the remaining as palettized, so don't fill
        // anything.
#ifdef DEBUG_BUILD
        printf("[GetBitMask] : Unknown FOURCC, might be palettized format.\n");
#endif
	*puiRmask = *puiGmask = *puiBmask = 0;
        *piBPP = 8;
        break;
  }
}


/////////////////////////////////////////////////////////////////////
//
// KillPM
//
// Simulates a switching to fullscreen mode,
// from the PM's point of view
//
/////////////////////////////////////////////////////////////////////
static void KillPM()
{
  HAB  hab;
  HWND hwnd;
  HDC  hdc;
  int  rc;

  hab = WinQueryAnchorBlock(HWND_DESKTOP);
  hwnd = WinQueryDesktopWindow(hab, 0);
  hdc = WinQueryWindowDC(hwnd);                    // Get HDC of desktop
  WinLockWindowUpdate(HWND_DESKTOP, HWND_DESKTOP); // Don't let other applications write to screen anymore!
  rc = GreDeath(hdc);                              // This is the standard way to tell the graphical engine
                                                   // that a switching to full-screen mode is about to come!
                                                   // (see gradd.inf)
}

/////////////////////////////////////////////////////////////////////
//
// RestorePM
//
// Simulates a switching back from fullscreen mode,
// from the PM's point of view
//
/////////////////////////////////////////////////////////////////////
static void RestorePM()
{
  HAB  hab;
  HWND hwnd;
  HDC  hdc;
  int  rc;

  hab = WinQueryAnchorBlock(HWND_DESKTOP);
  hwnd = WinQueryDesktopWindow(hab, 0);
  hdc = WinQueryWindowDC(hwnd);
  rc = GreResurrection(hdc, 0, NULL);      // This is the standard way of telling the graphical engine
                                           // that somebody has switched back from a fullscreen session to PM.
  WinLockWindowUpdate(HWND_DESKTOP, 0);    // Let others write to the screen again...
  WinInvalidateRect(hwnd, NULL, TRUE);     // Let everyone redraw itself! (Build the screen)
}

/////////////////////////////////////////////////////////////////////
//
// SetPointerVisibility
//
// Shows/hides the mouse pointer, using the VMAN
//
/////////////////////////////////////////////////////////////////////
static int SetPointerVisibility(SDL_VideoDevice *pVideo, BOOL fState)
{
  HWSHOWPTRIN hwspi;

  hwspi.ulLength=sizeof(hwspi);
  hwspi.fShow=fState;
  return (pVideo->hidden->pfnVMIEntry(0, VMI_CMD_SHOWPTR, &hwspi, NULL));
}

/////////////////////////////////////////////////////////////////////
//
// SwitchToFullscreen
//
// Switches to fullscreen-mode
//
/////////////////////////////////////////////////////////////////////
static int SwitchToFullscreen(SDL_VideoDevice *pVideo, PGDDMODEINFO OneMode)
{
  long ModeID;
  int rc;
  APERTURE DIVEaperture;
  FBINFO   DIVEfbinfo;

  // Switching to fullscreen goes this way:
  // - Tell PM (and the Graphical Engine GRE) that prepare
  //   for going to background (KillPM)
  // - Set new video mode using VMAN
  // - Tell DIVE that we're in FSDIVE mode from now
  //

  // Setup Aperture and FBINFO for FSDIVE
  memset(&DIVEfbinfo, 0, sizeof(FBINFO));

  DIVEfbinfo.ulLength = sizeof(FBINFO);
  DIVEfbinfo.ulCaps = 0;
  DIVEfbinfo.ulBPP  = OneMode->ulBpp;
  DIVEfbinfo.ulXRes = OneMode->ulHorizResolution;
  DIVEfbinfo.ulYRes = OneMode->ulVertResolution;
  DIVEfbinfo.ulScanLineBytes = OneMode->ulScanLineSize;
  DIVEfbinfo.ulNumENDIVEDrivers = 0; // unknown
  DIVEfbinfo.fccColorEncoding = OneMode->fccColorEncoding;

  DIVEaperture.ulPhysAddr = (long) OneMode->pbVRAMPhys;
  DIVEaperture.ulApertureSize = OneMode->ulApertureSize;
  DIVEaperture.ulScanLineSize = OneMode->ulScanLineSize;
  DIVEaperture.rctlScreen.yBottom = OneMode->ulVertResolution - 1;
  DIVEaperture.rctlScreen.xRight = OneMode->ulHorizResolution - 1;
  DIVEaperture.rctlScreen.yTop = 0;
  DIVEaperture.rctlScreen.xLeft = 0;

  SetPointerVisibility(pVideo, FALSE);             // Hide mouse
  KillPM();                                        // Tell PM that we're switching away!
  DosSleep(256);                                   // Give some time for it...
  ModeID = OneMode->ulModeId;
  rc = pVideo->hidden->pfnVMIEntry(0, VMI_CMD_SETMODE, &ModeID, NULL);// Set new video mode
  if (rc!=NO_ERROR)
  {
    // Rollback if something went wrong
    RestorePM();
    SetPointerVisibility(pVideo, TRUE);
  } else
  {
    DiveFullScreenInit(&DIVEaperture, &DIVEfbinfo); // Tell DIVE that it can work in this
    pVideo->hidden->fInFullScreenNow = TRUE;        // fullscreen mode now!
  }
  return (rc==NO_ERROR);
}

/////////////////////////////////////////////////////////////////////
//
// SwitchBackToDesktop
//
// Switches back from fullscreen-mode
//
/////////////////////////////////////////////////////////////////////
static int SwitchBackToDesktop(SDL_VideoDevice *pVideo)
{
  long ModeID;
  int rc;

  // Switching back from fullscreen mode:
  // - Set the old Desktop video mode
  // - Tell DIVE that it's not in FSDIVE mode anymore
  // - Tell the PM and the GRE that we're switched back

  ModeID = pVideo->hidden->DesktopModeInfo.ulModeId;
  rc = pVideo->hidden->pfnVMIEntry(0, VMI_CMD_SETMODE, &ModeID, NULL); // Set old video mode
  DosSleep(256);
  DiveFullScreenTerm();                             // Tell DIVE that end of fullscreen mode
  RestorePM();                                      // Tell PM that we're switching back
  pVideo->hidden->fInFullScreenNow = FALSE;
  SetPointerVisibility(pVideo, TRUE);                       // Restore mouse pointer

  return TRUE;
}

/////////////////////////////////////////////////////////////////////
//
// TranslateKey
//
// This creates SDL Keycodes from VK_ and hardware scan codes
//
/////////////////////////////////////////////////////////////////////
static SDL_keysym *TranslateKey(int vkey, int chcode, int scancode, SDL_keysym *keysym, int iPressed)
{
  keysym->scancode = (unsigned char) scancode;
  if (vkey!=0)
  // Translate keyboard code based on virtual key code, if possible!
    keysym->sym = VK_keymap[vkey];
  else
  // Use the ch code otherwise!
    keysym->sym = chcode & 0xFF;
  keysym->mod = KMOD_NONE;
  keysym->unicode = 0;
  if (iPressed && SDL_TranslateUNICODE)
  {
    // TODO:
    // Implement real unicode conversion!
    if (chcode)
      keysym->unicode = chcode;
    else
      keysym->unicode = vkey;
  }
#ifdef DEBUG_BUILD
  if (keysym->sym == SDLK_UNKNOWN)
  {
    printf("Unknown key: vkey = 0x%x  chcode=0x%x (%c)  scancode=0x%x\n",
           vkey, chcode, scancode); fflush(stdout);
  }
#endif
  return keysym;
}


/////////////////////////////////////////////////////////////////////
//
// WndProc
//
// This is the message processing window procedure for the
// SDLWindowClass, which is the client window in our application.
// It handles switching back and away from the app (taking care of
// going out and back to and from fullscreen mode), sending keystrokes
// and mouse events to where it has to be sent, etc...
//
/////////////////////////////////////////////////////////////////////
static MRESULT EXPENTRY WndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
  int active;
  HPS ps;
  SWP swp;
  RECTL rcl;
  HRGN rgn;
  RGNRECT rgnctl;
  POINTL pl;
  RECTL rcls[50];
  SETUP_BLITTER SetupBlitter;
  SDL_VideoDevice *pVideo = NULL;

  switch (msg)
  {
    case WM_SIZE:         // Window resizing
    case WM_MINMAXFRAME:  // Window minimize/maximize/restore
#ifdef DEBUG_BUILD
      printf(" WM_MINMAXFRAME\n"); fflush(stdout);
#endif
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
	if (pVideo->hidden->fFullScreenMode)
	{
	  return 0;         // do nothing if in fullscreen mode!
	}                   // else do the normal processing.
      }
      break;

    case WM_ACTIVATE:     // activation/deactivation of window
#ifdef DEBUG_BUILD
      printf("WM_ACTIVATE\n"); fflush(stdout);
#endif
      active = (int) mp1;

      if (!active)
      { // Switching away from the applicatoin
	pVideo = WinQueryWindowPtr(hwnd, 0);
	if (pVideo)
	{
	  if (pVideo->hidden->fInFullScreenNow)   // Switching away from the application, which is in FS mode!
	  {
	    SwitchBackToDesktop(pVideo);
	    WinShowWindow(pVideo->hidden->hwndFrame, FALSE);                // hide window
	    WinSetVisibleRegionNotify( pVideo->hidden->hwndClient, FALSE ); // turn VRN off
	    WinPostMsg(hwnd, WM_VRNDISABLED, 0, 0);       // make sure dive turns off itself too
	    WinSetCapture(HWND_DESKTOP, 0);               // release captured mouse
	  }
          pVideo->hidden->fInFocus = 0;
	}
      } else
      { // Switching to the application
	pVideo = WinQueryWindowPtr(hwnd, 0);
	if (pVideo)
	{
	  if (pVideo->hidden->fFullScreenMode)  // Switching to the application which should run in FS mode!
	  {
	    WinShowWindow(pVideo->hidden->hwndFrame, TRUE);                 // make window visible
	    WinSetVisibleRegionNotify( pVideo->hidden->hwndClient, TRUE );  // turn on VRN
	    WinPostMsg(hwnd, WM_VRNENABLED, 0, 0);        // setup dive
	    WinSetCapture(HWND_DESKTOP, pVideo->hidden->hwndClient);        // capture mouse
	    SwitchToFullscreen(pVideo, pVideo->hidden->FullScreenModeInfo); // and do switching!
	  }
          pVideo->hidden->fInFocus = 1;
	}
      }
      break;

    case WM_VRNDISABLED: // Visible region notification
#ifdef DEBUG_BUILD
      printf("WM_VRNDISABLED\n"); fflush(stdout);
#endif
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
	pVideo->hidden->fVRNDisabled = TRUE;
	// Dive must be protected by mutex semaphore if it can be accessed
	// from another thread too.
	if (DosRequestMutexSem(pVideo->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT) == NO_ERROR)
	{
	  DiveSetupBlitter(pVideo->hidden->hDiveInst, NULL);
	  DosReleaseMutexSem(pVideo->hidden->hmtxUseDive);
	}
	return (MRESULT) FALSE;
      }
      break;
    case WM_VRNENABLED:  // Visible region notification
#ifdef DEBUG_BUILD
      printf("WM_VRNENABLED\n"); fflush(stdout);
#endif
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        if (DosRequestMutexSem(pVideo->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT) == NO_ERROR)
	{
	  ps = WinGetPS (hwnd);
	  rgn = GpiCreateRegion(ps, 0, NULL);

	  WinQueryVisibleRegion(hwnd, rgn);

	  rgnctl.ircStart = 1;
	  rgnctl.crc = 50;
	  rgnctl.ulDirection = 1;

	  // Get the all ORed rectangles

	  GpiQueryRegionRects(ps, rgn, NULL, &rgnctl, rcls);
  
	  GpiDestroyRegion(ps, rgn);

	  WinReleasePS(ps);

	  // Now find the window position and size, relative to parent.

	  WinQueryWindowPos( hwnd, &swp );

	  // Convert the point to offset from desktop lower left.

	  pl.x = swp.x;
	  pl.y = swp.y;

	  WinMapWindowPoints(pVideo->hidden->hwndFrame, HWND_DESKTOP, &pl, 1 );

	  // Tell DIVE about the new settings.

	  SetupBlitter.ulStructLen       = sizeof( SETUP_BLITTER );
	  SetupBlitter.fccSrcColorFormat = pVideo->hidden->fccFourCC;
	  SetupBlitter.ulSrcWidth        = pVideo->hidden->iWidth;
	  SetupBlitter.ulSrcHeight       = pVideo->hidden->iHeight;
	  SetupBlitter.ulSrcPosX         = 0;
	  SetupBlitter.ulSrcPosY         = 0;
	  SetupBlitter.fInvert           = FALSE; // Don't invert, so the first line of buffer is the top-most line of image
	  SetupBlitter.ulDitherType      = 1;

	  SetupBlitter.fccDstColorFormat = FOURCC_SCRN;
	  SetupBlitter.lDstPosX          = 0;
	  SetupBlitter.lDstPosY          = 0;

	  if (!(pVideo->hidden->fFullScreenMode))
	  {
	    SetupBlitter.ulDstWidth        = swp.cx;
	    SetupBlitter.ulDstHeight       = swp.cy;

	    SetupBlitter.lScreenPosX       = pl.x;
	    SetupBlitter.lScreenPosY       = pl.y;

	    SetupBlitter.ulNumDstRects     = rgnctl.crcReturned;
	    SetupBlitter.pVisDstRects      = rcls;
	  } else
	  {
	    SetupBlitter.ulDstWidth        = pVideo->hidden->FullScreenModeInfo->ulHorizResolution;
	    SetupBlitter.ulDstHeight       = pVideo->hidden->FullScreenModeInfo->ulVertResolution;

	    SetupBlitter.lScreenPosX       = 0;
	    SetupBlitter.lScreenPosY       = 0;

	    SetupBlitter.ulNumDstRects     = DIVE_FULLY_VISIBLE;
	    SetupBlitter.pVisDstRects      = NULL;
	  }

	  DiveSetupBlitter(pVideo->hidden->hDiveInst, &SetupBlitter );
          pVideo->hidden->fVRNDisabled = FALSE;
	  DosReleaseMutexSem(pVideo->hidden->hmtxUseDive);
	}
      }
      return (MRESULT) FALSE;

    case WM_CHAR:  // Keypress notification
#ifdef DEBUG_BUILD
      printf("WM_CHAR\n"); fflush(stdout);
#endif
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
#ifdef DEBUG_BUILD
        // ----- Some private stuff for debugging
	if (SHORT1FROMMP ( mp2 ) == 'f')
        {
          printf(" FS Switch fastkey!\n"); fflush(stdout);
	  if (pVideo->hidden->fFullScreenMode)
	  { // Switch to windowed
	    SwitchBackToDesktop(pVideo);
	    WinSetCapture(HWND_DESKTOP, 0);
	  } else
	  { // Switch to full screen
	    SwitchToFullscreen(pVideo, pVideo->hidden->FullScreenModeInfo);
	    WinSetCapture(HWND_DESKTOP, pVideo->hidden->hwndClient);
	  }
	  WinShowWindow(pVideo->hidden->hwndFrame, TRUE);
	  WinSetVisibleRegionNotify(pVideo->hidden->hwndClient, TRUE );
	  WinPostMsg(hwnd, WM_VRNENABLED, 0, 0);
	  pVideo->hidden->fFullScreenMode = pVideo->hidden->fInFullScreenNow;
	}

        if (SHORT1FROMMP ( mp2 ) == 'q')
        {
          printf(" Quit fastkey!\n"); fflush(stdout);
          WinPostMsg(hwnd, WM_CLOSE, 0, 0);
        }
#endif
        // We don't take care of repeated keys:
        // TODO: solve this!
/*        if (CHARMSG(&msg)->cRepeat>1)
        {
#ifdef DEBUG_BUILD
          printf("Repeated key (%d), skipping...\n", CHARMSG(&msg)->cRepeat); fflush(stdout);
#endif
          return 0;
        }*/

        // If it's not repeated, then let's see if its pressed or released!
	if (SHORT1FROMMP(mp1) & KC_KEYUP)
	{
	  // A key has been released
          SDL_keysym keysym;
#ifdef DEBUG_BUILD
          printf(" Key UP!\n"); fflush(stdout);
#endif

	  SDL_PrivateKeyboard(SDL_RELEASED, TranslateKey(SHORT2FROMMP(mp2), // VK_ code
							 SHORT1FROMMP(mp2), // Character code
							 CHAR4FROMMP(mp1),  // HW Scan code
                                                         &keysym,0));
	} else
	{
          // A key has been pressed
          SDL_keysym keysym;
#ifdef DEBUG_BUILD
          printf(" Key DOWN!\n"); fflush(stdout);
#endif

	  SDL_PrivateKeyboard(SDL_PRESSED, TranslateKey(SHORT2FROMMP(mp2), // VK_ code
							SHORT1FROMMP(mp2), // Character code
							CHAR4FROMMP(mp1),  // HW Scan code
							&keysym,1));
	}
      }
      return 0;

    case WM_PAINT:  // Window redraw!
#ifdef DEBUG_BUILD
      printf("WM_PAINT\n"); fflush(stdout);
#endif
      ps = WinBeginPaint(hwnd,0,&rcl);
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        if ((pVideo->hidden->ulImage) && (!(pVideo->hidden->fVRNDisabled)))
	if (DosRequestMutexSem(pVideo->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT) == NO_ERROR)
	{
	  // TODO: We should only blit what has been changed

	  DiveBlitImage(pVideo->hidden->hDiveInst, pVideo->hidden->ulImage, DIVE_BUFFER_SCREEN );

	  DosReleaseMutexSem(pVideo->hidden->hmtxUseDive);
	}
      }
      WinEndPaint(ps);
      return 0;

    case WM_BUTTON1DOWN:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_PRESSED,
			       SDL_BUTTON_LEFT,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_BUTTON1UP:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_RELEASED,
			       SDL_BUTTON_LEFT,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_BUTTON2DOWN:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_PRESSED,
			       SDL_BUTTON_RIGHT,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_BUTTON2UP:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_RELEASED,
			       SDL_BUTTON_RIGHT,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_BUTTON3DOWN:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_PRESSED,
			       SDL_BUTTON_MIDDLE,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_BUTTON3UP:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseButton(SDL_RELEASED,
			       SDL_BUTTON_MIDDLE,
			       ppts->x,
			       swp.cy - ppts->y);
      }
      break;
    case WM_MOUSEMOVE:
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        POINTS *ppts = (POINTS *) (&mp1);
        SWP swp;
        // We have to inverse the mouse position, because every non-os/2 system
        // has a coordinate system where the (0;0) is the top-left corner,
        // while on os/2 it's the bottom left corner!
        WinQueryWindowPos(hwnd, &swp);
	SDL_PrivateMouseMotion(0, // Buttons not changed
			       0, // Absolute position
			       ppts->x,
                               swp.cy - ppts->y);
      }
      break;
    case WM_CLOSE: // Window close
#ifdef DEBUG_BUILD
      printf("WM_CLOSE\n"); fflush(stdout);
#endif
      pVideo = WinQueryWindowPtr(hwnd, 0);
      if (pVideo)
      {
        // Check if it's ok to quit the application!
	if (SDL_PrivateQuit())
	{
	  if (pVideo->hidden->fFullScreenMode)
	  { // Switch to windowed
	    SwitchBackToDesktop(pVideo);
	    WinSetCapture(HWND_DESKTOP, 0);
	    WinShowWindow(pVideo->hidden->hwndFrame, TRUE);
	    WinSetVisibleRegionNotify(pVideo->hidden->hwndClient, TRUE );
	    WinPostMsg(hwnd, WM_VRNENABLED, 0, 0);
	    pVideo->hidden->fFullScreenMode = pVideo->hidden->fInFullScreenNow;
	  }
	  WinPostMsg (hwnd, WM_QUIT, 0, 0);
	} else return 0;
      }
      break;

    default:
      break;
  }
  // Run the default window procedure for unhandled stuffs
  return WinDefWindowProc(hwnd, msg, mp1, mp2);
}

/////////////////////////////////////////////////////////////////////
//
// PMThreadFunc
//
// This function implements the PM-Thread, which initializes the
// application window itself, the DIVE, and start message processing.
//
/////////////////////////////////////////////////////////////////////
static void PMThreadFunc(void *pParm)
{
  SDL_VideoDevice *pVideo = pParm;
  HAB hab;
  HMQ hmq;
  QMSG msg;
  ULONG fcf;
  unsigned int uiRmask, uiGmask, uiBmask;
  int iBPP;
  int iModeNum;
  PGDDMODEINFO pModeInfo;

#ifdef DEBUG_BUILD
  printf("[PMThreadFunc] : Starting\n"); fflush(stdout);
#endif

  // Initialize PM, create a message queue.

  hab=WinInitialize(0);
  hmq=WinCreateMsgQueue(hab,0);
  if (hmq==0)
  {
#ifdef DEBUG_BUILD
    printf("[PMThreadFunc] : Could not create message queue!\n");
    printf("                 It might be that the application using SDL is not a PM app!\n");
    fflush(stdout);
#endif
    pVideo->hidden->iPMThreadStatus = 2;
  } else
  {
    // Now check the capabilities of Dive!
    if (!InspectDIVE(pVideo))
    {
#ifdef DEBUG_BUILD
      printf("[PMThreadFunc] : Could not get Dive capabilities!\n");
      fflush(stdout);
#endif
      pVideo->hidden->iPMThreadStatus = 3;
    } else
    {
  
      // Create mutex semaphore to protect the DIVE
      DosCreateMutexSem(NULL, &(pVideo->hidden->hmtxUseDive), 0, FALSE);

      // Remember, that the application is NOT in fullscreen mode
      // by default, and we will be in focus
      pVideo->hidden->fFullScreenMode=FALSE;
      pVideo->hidden->fInFocus = 1;

      // Register our private window class with our window procedure
      WinRegisterClass( hab, "SDLWindowClass", WndProc, CS_SIZEREDRAW, 8);
      fcf =
	FCF_TITLEBAR |
	FCF_SYSMENU |
	FCF_MINBUTTON |
	FCF_MAXBUTTON |
	FCF_NOBYTEALIGN |
	FCF_SIZEBORDER |
        FCF_TASKLIST;

      pVideo->hidden->hwndFrame = WinCreateStdWindow( HWND_DESKTOP, 0, &fcf,
						      "SDLWindowClass",
						      "SDL Application", WS_VISIBLE, //WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
						      0, 0, &(pVideo->hidden->hwndClient));

      // Store pVideo pointer in window data for client window, so
      // it will know the instance to which it belongs to.
      WinSetWindowPtr(pVideo->hidden->hwndClient, 0, pVideo);

      // Now set default image width height and fourcc!

      // The default image width and height will be the first available
      // fullscreen video mode which is not 8BPP!
      pModeInfo = pVideo->hidden->ModeInfo; // Start from the first available FS mode
      for (iModeNum=0; iModeNum<pVideo->hidden->NumVideoModes; iModeNum++)
      {
        // Go through all the modes, and exit at the first video mode which is not 8bpp
        if (pModeInfo->ulBpp>8) break;
        pModeInfo++;
      }
      // If we haven't found such a mode, then use the first one!
      if (iModeNum>=pVideo->hidden->NumVideoModes)
        pModeInfo = pVideo->hidden->ModeInfo;

      // Ok, pModeInfo points to a good FullScreen mode descriptor.
      // Now we'll set up DIVE to such a buffer, but we have to take care to
      // create such a buffer, that the dive can use that as an input buffer to
      // be blitted to the screen. So, take care to only use FourCCs what dive accepts
      // as input buffer!
      pVideo->hidden->fccFourCC = NearestAcceptableDiveInputFCC(pVideo, pModeInfo->fccColorEncoding);
      // Get the properties of this FourCC (we're only interested in iBPP now)
      GetBitMask(pVideo->hidden->fccFourCC,
                 &uiRmask, &uiGmask, &uiBmask, &iBPP);

      // Dive buffer's width and height will be this:
      pVideo->hidden->iWidth = pModeInfo->ulHorizResolution;
      pVideo->hidden->iHeight = pModeInfo->ulVertResolution;

      // Set this fulscreen mode to be the one what we'll switch to when the
      // user asks for it!
      pVideo->hidden->FullScreenModeInfo = pModeInfo;

      // Open Dive, allocate image buffer,
      // and set VRN, and setup Dive Blitter for
      // the first time (WM_VRNENABLED msg.)
      pVideo->hidden->fVRNDisabled = TRUE;
  
      DiveOpen( &(pVideo->hidden->hDiveInst), FALSE, 0 );
      // Calculate scan line size
      pVideo->hidden->ulScanLineBytes = pVideo->hidden->iWidth * ((iBPP+7)/8);
      // Make it a multiple of 8 bytes (needed by SDL??)
      pVideo->hidden->ulScanLineBytes = ((pVideo->hidden->ulScanLineBytes+7)/8)*8;
      pVideo->hidden->pBuffer = (void *) malloc(pVideo->hidden->ulScanLineBytes * pVideo->hidden->iHeight);

      if (!(pVideo->hidden->pBuffer))
      {
	// Not enough memory for video buffer!!!!
        DiveClose( pVideo->hidden->hDiveInst ); pVideo->hidden->hDiveInst= NULL;
  
	// Destroy our window
	WinDestroyWindow(pVideo->hidden->hwndFrame); pVideo->hidden->hwndFrame=NULL;

	// Destroy Dive protector semaphore
	DosReleaseMutexSem(pVideo->hidden->hmtxUseDive);
	DosCloseMutexSem(pVideo->hidden->hmtxUseDive);  pVideo->hidden->hmtxUseDive= NULL;

	// Free result of InspectDIVE()!
	free(pVideo->hidden->pfccFormats);
	free(pVideo->hidden->pDiveCaps);

	// Uninitialize PM
	WinDestroyMsgQueue(hmq);
	WinTerminate(hab);

#ifdef DEBUG_BUILD
	printf("[PMThreadFunc] : Out of memory allocating video buffer!\n");
	fflush(stdout);
#endif
	pVideo->hidden->iPMThreadStatus = 4;
      } else
      {
        // Ok, we could allocate the memory. Good.
        // Get a dive buffe handle from this
	DiveAllocImageBuffer( pVideo->hidden->hDiveInst, &(pVideo->hidden->ulImage),
			      pVideo->hidden->fccFourCC,
			      pVideo->hidden->iWidth, pVideo->hidden->iHeight,
			      pVideo->hidden->ulScanLineBytes, pVideo->hidden->pBuffer);

	// Start VisibleRegionNotify messages to be sent to our window
        // (so we can change DIVE blitter every time the VR changes)
	WinSetVisibleRegionNotify( pVideo->hidden->hwndClient, TRUE );
	// Make sure the DIVE blitter will be set up at least once:
	WinPostMsg( pVideo->hidden->hwndClient, WM_VRNENABLED, 0, 0 );

	// Set the position and size of the main window,
        // and make it visible!
	SetAccessableWindowPos(pVideo->hidden->hwndFrame,
			       HWND_TOP,
			       (WinQuerySysValue (HWND_DESKTOP, SV_CXSCREEN) - pVideo->hidden->iWidth) / 2,
			       (WinQuerySysValue (HWND_DESKTOP, SV_CYSCREEN) - pVideo->hidden->iHeight) / 2,
			       pVideo->hidden->iWidth + WinQuerySysValue (HWND_DESKTOP, SV_CXDLGFRAME)*2,
			       pVideo->hidden->iHeight + WinQuerySysValue (HWND_DESKTOP, SV_CYTITLEBAR)
			       + WinQuerySysValue (HWND_DESKTOP, SV_CYDLGFRAME)*2,
			       SWP_SIZE | SWP_ACTIVATE | SWP_SHOW | SWP_MOVE);

#ifdef DEBUG_BUILD
	printf("[PMThreadFunc] : Entering message loop\n"); fflush(stdout);
#endif
	pVideo->hidden->iPMThreadStatus = 1;
  
	while (WinGetMsg(hab, (PQMSG)&msg, 0, 0, 0))
	  WinDispatchMsg(hab, (PQMSG) &msg);

#ifdef DEBUG_BUILD
	printf("[PMThreadFunc] : Leaving message loop\n"); fflush(stdout);
#endif
  
	// Shut down DIVE
	// We must switch back to windowed mode before closing Dive!
	// (Because switching uses Dive too!)
	if (pVideo->hidden->fInFullScreenNow)
	  SwitchBackToDesktop(pVideo);

        // Stop VRN messages
	WinSetVisibleRegionNotify( pVideo->hidden->hwndClient, FALSE );

        // Free image buffer and close DIVE
	DosRequestMutexSem(pVideo->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT);
	DiveFreeImageBuffer( pVideo->hidden->hDiveInst, pVideo->hidden->ulImage );
	pVideo->hidden->ulImage = 0;
	if (pVideo->hidden->pBuffer)
	{
	  free(pVideo->hidden->pBuffer); pVideo->hidden->pBuffer = NULL;
	}
	DiveClose( pVideo->hidden->hDiveInst ); pVideo->hidden->hDiveInst= NULL;
  
	// Destroy our window
	WinDestroyWindow(pVideo->hidden->hwndFrame); pVideo->hidden->hwndFrame=NULL;
  
	// Destroy Dive protector semaphore
	DosReleaseMutexSem(pVideo->hidden->hmtxUseDive);
	DosCloseMutexSem(pVideo->hidden->hmtxUseDive);  pVideo->hidden->hmtxUseDive= NULL;
  
	// Free result of InspectDIVE()!
	free(pVideo->hidden->pfccFormats);
	free(pVideo->hidden->pDiveCaps);

	// Uninitialize PM
	WinDestroyMsgQueue(hmq);
	WinTerminate(hab);
  
	// All done!
	pVideo->hidden->iPMThreadStatus = 0;
      }
    }
  }
#ifdef DEBUG_BUILD
  printf("[PMThreadFunc] : End, status is %d!\n", pVideo->hidden->iPMThreadStatus); fflush(stdout);
#endif
  _endthread();
}

/// ---------------- DUMMY FUNCTIONS --------------------------

/* Free a window manager cursor */
void os2video_FreeWMCursor(_THIS, WMcursor *cursor)
{
}

/* Create a black/white window manager cursor */
WMcursor *os2video_CreateWMCursor(_THIS, Uint8 *data, Uint8 *mask,
                                  int w, int h, int hot_x, int hot_y)
{
  return NULL;
}

/* Show the specified cursor, or hide if cursor is NULL */
int os2video_ShowWMCursor(_THIS, WMcursor *cursor)
{
  return 0;
}

/* Warp the window manager cursor to (x,y)
 If NULL, a mouse motion event is posted internally.
 */
void os2video_WarpWMCursor(_THIS, Uint16 x, Uint16 y)
{
}

/* If not NULL, this is called when a mouse motion event occurs */
void os2video_MoveWMCursor(_THIS, int x, int y)
{
}

/* Determine whether the mouse should be in relative mode or not.
 This function is called when the input grab state or cursor
 visibility state changes.
 If the cursor is not visible, and the input is grabbed, the
 driver can place the mouse in relative mode, which may result
 in higher accuracy sampling of the pointer motion.
 */
void os2video_CheckMouseMode(_THIS)
{
}

static void os2video_PumpEvents(_THIS)
{
  // We have everything in our PM thread, so no need for PumpEvents()...
}

static void os2video_InitOSKeymap(_THIS)
{
  int i;

  /* Map the VK and CH keysyms */
  for ( i=0; i<SDL_TABLESIZE(VK_keymap); ++i )
    VK_keymap[i] = SDLK_UNKNOWN;

  // Tested:

  VK_keymap[VK_ESC] = SDLK_ESCAPE;
  VK_keymap[VK_F1] = SDLK_F1;
  VK_keymap[VK_F2] = SDLK_F2;
  VK_keymap[VK_F3] = SDLK_F3;
  VK_keymap[VK_F4] = SDLK_F4;
  VK_keymap[VK_F5] = SDLK_F5;
  VK_keymap[VK_F6] = SDLK_F6;
  VK_keymap[VK_F7] = SDLK_F7;
  VK_keymap[VK_F8] = SDLK_F8;
  VK_keymap[VK_F9] = SDLK_F9;
  VK_keymap[VK_F10] = SDLK_F10;
  VK_keymap[VK_F11] = SDLK_F11;
  VK_keymap[VK_F12] = SDLK_F12;
  VK_keymap[VK_F13] = SDLK_F13;
  VK_keymap[VK_F14] = SDLK_F14;
  VK_keymap[VK_F15] = SDLK_F15;

  VK_keymap[VK_BACKSPACE] = SDLK_BACKSPACE;

  VK_keymap[VK_TAB] = SDLK_TAB;


  // Untested:


  VK_keymap[VK_CLEAR] = SDLK_CLEAR;
  VK_keymap[VK_NEWLINE] = SDLK_RETURN;
  VK_keymap[VK_PAUSE] = SDLK_PAUSE;
  VK_keymap[VK_ESC] = SDLK_ESCAPE;
  VK_keymap[VK_SPACE] = SDLK_SPACE;
  VK_keymap[VK_APOSTROPHE] = SDLK_QUOTE;
  VK_keymap[VK_COMMA] = SDLK_COMMA;
  VK_keymap[VK_MINUS] = SDLK_MINUS;
  VK_keymap[VK_PERIOD] = SDLK_PERIOD;
  VK_keymap[VK_SLASH] = SDLK_SLASH;
  VK_keymap[VK_0] = SDLK_0;
  VK_keymap[VK_1] = SDLK_1;
  VK_keymap[VK_2] = SDLK_2;
  VK_keymap[VK_3] = SDLK_3;
  VK_keymap[VK_4] = SDLK_4;
  VK_keymap[VK_5] = SDLK_5;
  VK_keymap[VK_6] = SDLK_6;
  VK_keymap[VK_7] = SDLK_7;
  VK_keymap[VK_8] = SDLK_8;
  VK_keymap[VK_9] = SDLK_9;
  VK_keymap[VK_SEMICOLON] = SDLK_SEMICOLON;
  VK_keymap[VK_EQUALS] = SDLK_EQUALS;
  VK_keymap[VK_LBRACKET] = SDLK_LEFTBRACKET;
  VK_keymap[VK_BACKSLASH] = SDLK_BACKSLASH;
  VK_keymap[VK_RBRACKET] = SDLK_RIGHTBRACKET;
  VK_keymap[VK_GRAVE] = SDLK_BACKQUOTE;
  VK_keymap[VK_BACKTICK] = SDLK_BACKQUOTE;
  VK_keymap[VK_A] = SDLK_a;
  VK_keymap[VK_B] = SDLK_b;
  VK_keymap[VK_C] = SDLK_c;
  VK_keymap[VK_D] = SDLK_d;
  VK_keymap[VK_E] = SDLK_e;
  VK_keymap[VK_F] = SDLK_f;
  VK_keymap[VK_G] = SDLK_g;
  VK_keymap[VK_H] = SDLK_h;
  VK_keymap[VK_I] = SDLK_i;
  VK_keymap[VK_J] = SDLK_j;
  VK_keymap[VK_K] = SDLK_k;
  VK_keymap[VK_L] = SDLK_l;
  VK_keymap[VK_M] = SDLK_m;
  VK_keymap[VK_N] = SDLK_n;
  VK_keymap[VK_O] = SDLK_o;
  VK_keymap[VK_P] = SDLK_p;
  VK_keymap[VK_Q] = SDLK_q;
  VK_keymap[VK_R] = SDLK_r;
  VK_keymap[VK_S] = SDLK_s;
  VK_keymap[VK_T] = SDLK_t;
  VK_keymap[VK_U] = SDLK_u;
  VK_keymap[VK_V] = SDLK_v;
  VK_keymap[VK_W] = SDLK_w;
  VK_keymap[VK_X] = SDLK_x;
  VK_keymap[VK_Y] = SDLK_y;
  VK_keymap[VK_Z] = SDLK_z;
  VK_keymap[VK_DELETE] = SDLK_DELETE;

//  VK_keymap[VK_NUMPAD0] = SDLK_KP0;
//  VK_keymap[VK_NUMPAD1] = SDLK_KP1;
//  VK_keymap[VK_NUMPAD2] = SDLK_KP2;
//  VK_keymap[VK_NUMPAD3] = SDLK_KP3;
//  VK_keymap[VK_NUMPAD4] = SDLK_KP4;
//  VK_keymap[VK_NUMPAD5] = SDLK_KP5;
//  VK_keymap[VK_NUMPAD6] = SDLK_KP6;
//  VK_keymap[VK_NUMPAD7] = SDLK_KP7;
//  VK_keymap[VK_NUMPAD8] = SDLK_KP8;
//  VK_keymap[VK_NUMPAD9] = SDLK_KP9;
//  VK_keymap[VK_DECIMAL] = SDLK_KP_PERIOD;
//  VK_keymap[VK_DIVIDE] = SDLK_KP_DIVIDE;
//  VK_keymap[VK_MULTIPLY] = SDLK_KP_MULTIPLY;
//  VK_keymap[VK_SUBTRACT] = SDLK_KP_MINUS;
//  VK_keymap[VK_ADD] = SDLK_KP_PLUS;

  VK_keymap[VK_UP] = SDLK_UP;
  VK_keymap[VK_DOWN] = SDLK_DOWN;
  VK_keymap[VK_RIGHT] = SDLK_RIGHT;
  VK_keymap[VK_LEFT] = SDLK_LEFT;
  VK_keymap[VK_INSERT] = SDLK_INSERT;
  VK_keymap[VK_HOME] = SDLK_HOME;
  VK_keymap[VK_END] = SDLK_END;
//  VK_keymap[VK_PRIOR] = SDLK_PAGEUP;
//  VK_keymap[VK_NEXT] = SDLK_PAGEDOWN;

  VK_keymap[VK_NUMLOCK] = SDLK_NUMLOCK;
//  VK_keymap[VK_CAPITAL] = SDLK_CAPSLOCK;
//  VK_keymap[VK_SCROLL] = SDLK_SCROLLOCK;
//  VK_keymap[VK_RSHIFT] = SDLK_RSHIFT;
//  VK_keymap[VK_LSHIFT] = SDLK_LSHIFT;
//  VK_keymap[VK_RCONTROL] = SDLK_RCTRL;
//  VK_keymap[VK_LCONTROL] = SDLK_LCTRL;
//  VK_keymap[VK_RMENU] = SDLK_RALT;
//  VK_keymap[VK_LMENU] = SDLK_LALT;
//  VK_keymap[VK_RWIN] = SDLK_RSUPER;
//  VK_keymap[VK_LWIN] = SDLK_LSUPER;

//  VK_keymap[VK_HELP] = SDLK_HELP;
#ifdef VK_PRINT
  VK_keymap[VK_PRINT] = SDLK_PRINT;
#endif
//  VK_keymap[VK_SNAPSHOT] = SDLK_PRINT;
//  VK_keymap[VK_CANCEL] = SDLK_BREAK;
//  VK_keymap[VK_APPS] = SDLK_MENU;
}


/* We don't actually allow hardware surfaces other than the main one */
static int os2video_AllocHWSurface(_THIS, SDL_Surface *surface)
{
  return(-1);
}
static void os2video_FreeHWSurface(_THIS, SDL_Surface *surface)
{
  return;
}

/* We need to wait for vertical retrace on page flipped displays */
static int os2video_LockHWSurface(_THIS, SDL_Surface *surface)
{
  return(0);
}

static void os2video_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
  return;
}

static int os2video_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
  printf("os2video_SetColors\n"); fflush(stdout);
  // TODO: Implement paletted modes
  return(1);
}

/* Set the window icon image */
void os2video_SetIcon(_THIS, SDL_Surface *icon, Uint8 *mask)
{
  printf("os2video_SetIcon\n"); fflush(stdout);
  // TODO: Implement icons
}

// ------------------------ REAL FUNCTIONS -----------------


/* Iconify the window.
 This function returns 1 if there is a window manager and the
 window was actually iconified, it returns 0 otherwise.
 */
int os2video_IconifyWindow(_THIS)
{
  HAB hab;
  HMQ hmq;
  ERRORID hmqerror;

  // If there is no more window, nothing we can do!
  if (_this->hidden->iPMThreadStatus!=1) return 0;

  // Cannot do anything in fullscreen mode!
  if (_this->hidden->fInFullScreenNow)
    return 0;

  // Make sure this thread is prepared for using the Presentation Manager!
  hab = WinInitialize(0);
  hmq = WinCreateMsgQueue(hab,0);
  // Remember if there was an error at WinCreateMsgQueue(), because we don't
  // want to destroy somebody else's queue later. :)
  hmqerror = WinGetLastError(hab);

  WinSetWindowPos(_this->hidden->hwndFrame, HWND_TOP,
		 0, 0, 0, 0, SWP_MINIMIZE);

  // Now destroy the message queue, if we've created it!
  if (ERRORIDERROR(hmqerror)==0)
    WinDestroyMsgQueue(hmq);

  return 1;
}

/* Set the title and icon text */
static void os2video_SetCaption(_THIS, const char *title, const char *icon)
{
  HAB hab;
  HMQ hmq;
  ERRORID hmqerror;

  // If there is no more window, nothing we can do!
  if (_this->hidden->iPMThreadStatus!=1) return;

  // Make sure this thread is prepared for using the Presentation Manager!
  hab = WinInitialize(0);
  hmq = WinCreateMsgQueue(hab,0);
  // Remember if there was an error at WinCreateMsgQueue(), because we don't
  // want to destroy somebody else's queue later. :)
  hmqerror = WinGetLastError(hab);

  WinSetWindowText(_this->hidden->hwndFrame, (char *) title);

  // Now destroy the message queue, if we've created it!
  if (ERRORIDERROR(hmqerror)==0)
    WinDestroyMsgQueue(hmq);
}

static int os2video_ToggleFullScreen(_THIS, int on)
{
#ifdef DEBUG_BUILD
  printf("[os2video_ToggleFullScreen] : %d\n", on); fflush(stdout);
#endif

  // Check if it's really changing state!
  if (_this->hidden->fFullScreenMode!=on)
  {
    _this->hidden->fFullScreenMode = on;
#ifdef DEBUG_BUILD
    printf("FullScreenMode = %d\n", on); fflush(stdout);
#endif
    if (_this->hidden->fInFocus)
    {
#ifdef DEBUG_BUILD
      printf("Telling it to the hardware tood\n"); fflush(stdout);
#endif
      if (on)
      {
	SwitchToFullscreen(_this, _this->hidden->FullScreenModeInfo);
        if (_this->screen)
	  _this->screen->flags |= SDL_FULLSCREEN;
      } else
      {
	SwitchBackToDesktop(_this);
        if (_this->screen)
	  _this->screen->flags &= (~SDL_FULLSCREEN);
      }
    }

  }
  return 1;
}

/* This is called after the video mode has been set, to get the
 initial mouse state.  It should queue events as necessary to
 properly represent the current mouse focus and position.
 */
static void os2video_UpdateMouse(_THIS)
{
  POINTL ptl;
  HAB hab;
  HMQ hmq;
  ERRORID hmqerror;

  // If there is no more window, nothing we can do!
  if (_this->hidden->iPMThreadStatus!=1) return;

  // Make sure this thread is prepared for using the Presentation Manager!
  hab = WinInitialize(0);
  hmq = WinCreateMsgQueue(hab,0);
  // Remember if there was an error at WinCreateMsgQueue(), because we don't
  // want to destroy somebody else's queue later. :)
  hmqerror = WinGetLastError(hab);

  if (_this->hidden->fInFocus)
  {
    // If our app is in focus
    SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
    SDL_PrivateAppActive(1, SDL_APPINPUTFOCUS);
    SDL_PrivateAppActive(1, SDL_APPACTIVE);
    WinQueryPointerPos(HWND_DESKTOP, &ptl);
    WinMapWindowPoints(HWND_DESKTOP, _this->hidden->hwndClient, &ptl, 1);
    SDL_PrivateMouseMotion(0, 0, (Sint16) (ptl.x), (Sint16) (ptl.y));
  } else
  {
    // If we're not in focus
    SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
    SDL_PrivateAppActive(0, SDL_APPINPUTFOCUS);
    SDL_PrivateAppActive(0, SDL_APPACTIVE);
    SDL_PrivateMouseMotion(0, 0, (Sint16) -1, (Sint16) -1);
  }

  // Now destroy the message queue, if we've created it!
  if (ERRORIDERROR(hmqerror)==0)
    WinDestroyMsgQueue(hmq);
}

/* This pointer should exist in the native video subsystem and should
 point to an appropriate update function for the current video mode
 */
static void os2video_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
  // If there is no more window, nothing we can do!
  if (_this->hidden->iPMThreadStatus!=1) return;

  if ((_this->hidden->ulImage) && (!(_this->hidden->fVRNDisabled)))
  {
    if (DosRequestMutexSem(_this->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT) == NO_ERROR)
    {
      // TODO: We should only blit what has been changed
      DiveBlitImage(_this->hidden->hDiveInst, _this->hidden->ulImage, DIVE_BUFFER_SCREEN );
      DosReleaseMutexSem(_this->hidden->hmtxUseDive);
    }
  }
}


/* Reverse the effects VideoInit() -- called if VideoInit() fails
 or if the application is shutting down the video subsystem.
 */
static void os2video_VideoQuit(_THIS)
{
  int rc;

#ifdef DEBUG_BUILD
  printf("[os2video_VideoQuit]\n"); fflush(stdout);
#endif
  // Close PM stuff if running!
  if (_this->hidden->iPMThreadStatus == 1)
  {
    WinPostMsg(_this->hidden->hwndFrame, WM_CLOSE, (MPARAM) 0, (MPARAM) 0);
    DosWaitThread(&(_this->hidden->tidPMThread), DCWW_WAIT);
  }

  // Free result of an old ListModes() call, because there is
  // no FreeListModes() call in SDL!
  if (_this->hidden->pListModesResult)
  {
    free(_this->hidden->pListModesResult); _this->hidden->pListModesResult = NULL;
  }

  // Free list of available fullscreen modes
  if (_this->hidden->ModeInfo)
  {
    free(_this->hidden->ModeInfo); _this->hidden->ModeInfo = NULL;
  }

  // Send a TERMPROC command to VMAN, so inform it about
  // that this process has stopped using its services
  rc = _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
#ifdef DEBUG_BUILD
  if (rc!=NO_ERROR)
  {
    printf("[os2video_VideoQuit] : VMI_CMD_TERMPROC rc = 0x%x\n", rc);
    fflush(stdout);
  }
#endif

  // Unload VMAN DLL
  DosFreeModule(_this->hidden->hmodVMAN);
}

/* Set the requested video mode, returning a surface which will be
 set to the SDL_VideoSurface.  The width and height will already
 be verified by ListModes(), and the video subsystem is free to
 set the mode to a supported bit depth different from the one
 specified -- the desired bpp will be emulated with a shadow
 surface if necessary.  If a new mode is returned, this function
 should take care of cleaning up the current mode.
 */
static SDL_Surface *os2video_SetVideoMode(_THIS, SDL_Surface *current,
					  int width, int height, int bpp, Uint32 flags)
{
  PGDDMODEINFO pModeInfo, pModeInfoFound;
  unsigned int uiRmask, uiGmask, uiBmask;
  int iBPP;
  int iModeNum;
  HAB hab;
  HMQ hmq;
  ERRORID hmqerror;

  // If there is no more window, nothing we can do!
  if (_this->hidden->iPMThreadStatus!=1) return NULL;

  // TODO: Implement palette modes!
  if (bpp==8) bpp=32;

  // First check if there is such a video mode they want!
  pModeInfoFound = NULL;
#ifdef ALLOW_EVERY_MODE_IN_WINDOWED
  // Check if they need fullscreen or non-fullscreen video modes!
  if ((flags & SDL_FULLSCREEN) == 0)

  {
    // For windowed mode we support every resolution!
    pModeInfoFound = _this->hidden->ModeInfo;
  } else
#endif
  {
    // For fullscreen mode we don't support every resolution!
    // So, go through the video modes, and check for such a resolution!
    pModeInfo = _this->hidden->ModeInfo;
    for (iModeNum=0; iModeNum<_this->hidden->NumVideoModes; iModeNum++)
    {
      // Check all available fullscreen modes for this resolution
      if ((pModeInfo->ulHorizResolution == width) &&
	  (pModeInfo->ulVertResolution == height))
      {
	// If good resolution, try to find the exact BPP, or at least
        // something similar...
	if (!pModeInfoFound) pModeInfoFound = pModeInfo;
        else
	if (pModeInfoFound->ulBpp!=bpp)
          pModeInfoFound = pModeInfo;
      }
      pModeInfo++;
    }
  }

  // If we did not find a good fullscreen mode, then return NULL
  if (!pModeInfoFound) return NULL;

#ifdef DEBUG_BUILD
  printf("Found mode!\n"); fflush(stdout);
  PrintModeInfo(pModeInfoFound);
#endif

  // We can try create new surface!

  // Make sure this thread is prepared for using the Presentation Manager!
  hab = WinInitialize(0);
  hmq = WinCreateMsgQueue(hab,0);
  // Remember if there was an error at WinCreateMsgQueue(), because we don't
  // want to destroy somebody else's queue later. :)
  hmqerror = WinGetLastError(hab);

  // Come back from fullscreen mode, then
  // free old DIVE buffer
  if (_this->hidden->fInFullScreenNow)
    SwitchBackToDesktop(_this);

  // Stop sending VRNs
  WinSetVisibleRegionNotify( _this->hidden->hwndClient, FALSE );

  if (DosRequestMutexSem(_this->hidden->hmtxUseDive, SEM_INDEFINITE_WAIT)==NO_ERROR)
  {
    // Free old dive buffer
    if (_this->hidden->ulImage)
    {
      DiveFreeImageBuffer( _this->hidden->hDiveInst, _this->hidden->ulImage ); _this->hidden->ulImage = 0;
      if (_this->hidden->pBuffer)
	free(_this->hidden->pBuffer);
    }

    // Allocate new, but again, take care to make it such that dive will accept it
    // as an input fourcc format!
    _this->hidden->fccFourCC = NearestAcceptableDiveInputFCC(_this, pModeInfoFound->fccColorEncoding);
    // Get parameters for this (new) fourCC
    GetBitMask(_this->hidden->fccFourCC,
             &uiRmask, &uiGmask, &uiBmask, &iBPP);

    // Set new dive buffer parameters!
    _this->hidden->iWidth = width;
    _this->hidden->iHeight = height;

    // Calculate scan line size
    _this->hidden->ulScanLineBytes = _this->hidden->iWidth * ((iBPP+7)/8);
    // Make it a multiple of 8 bytes (neede by SDL??)
    _this->hidden->ulScanLineBytes = ((_this->hidden->ulScanLineBytes+7)/8)*8;
    // Allocate new video buffer!
    _this->hidden->pBuffer = (void *) malloc(_this->hidden->ulScanLineBytes * _this->hidden->iHeight);

    if (!_this->hidden->pBuffer)
    {
#ifdef DEBUG_BUILD
      printf("Yikes, not enough memory for new video buffer!\n"); fflush(stdout);
#endif
      DosReleaseMutexSem(_this->hidden->hmtxUseDive);
      return NULL;
    }

    // Take a buffer number from DIVE
    DiveAllocImageBuffer( _this->hidden->hDiveInst, &(_this->hidden->ulImage),
			  _this->hidden->fccFourCC,
			  _this->hidden->iWidth, _this->hidden->iHeight,
			  _this->hidden->ulScanLineBytes, _this->hidden->pBuffer);

    // Save fullscreen mode what we'll use
    _this->hidden->FullScreenModeInfo = pModeInfoFound;
    // We don't use DIVE anymore
    DosReleaseMutexSem(_this->hidden->hmtxUseDive);

    // Set the new size of the main window
    SetAccessableWindowPos( _this->hidden->hwndFrame,
			    HWND_TOP,
			    0, 0,
			    _this->hidden->iWidth + WinQuerySysValue (HWND_DESKTOP, SV_CXDLGFRAME)*2,
			    _this->hidden->iHeight + WinQuerySysValue (HWND_DESKTOP, SV_CYTITLEBAR)
			    + WinQuerySysValue (HWND_DESKTOP, SV_CYDLGFRAME)*2,
			    SWP_SIZE);

    if (_this->hidden->fInFocus)
    {
      // Notify window about new stuffs
      WinSetVisibleRegionNotify(_this->hidden->hwndClient, TRUE );
      WinPostMsg(_this->hidden->hwndClient, WM_VRNENABLED, 0, 0 );
    }
  }

  current->flags = 0;
  current->flags |= (flags & SDL_SWSURFACE);
  current->flags |= (flags & SDL_HWSURFACE);
  current->flags |= (flags & SDL_FULLSCREEN);
  current->flags |= SDL_PREALLOC;

  current->w = width;
  current->h = height;

  current->format->Rmask = uiRmask;
  current->format->Gmask = uiGmask;
  current->format->Bmask = uiBmask;
  current->format->BitsPerPixel = iBPP;
  current->format->BytesPerPixel = (iBPP+7)/8;

  // Set fullscreen mode flag, and switch to fullscreen if needed!
  if (flags & SDL_FULLSCREEN)
  {
#ifdef DEBUG_BUILD
    printf("Also trying to switch to fullscreen (InFocus = %d)\n", _this->hidden->fInFocus);
    fflush(stdout);
#endif
    _this->hidden->fFullScreenMode = 1;
    if (_this->hidden->fInFocus)
      SwitchToFullscreen(_this, _this->hidden->FullScreenModeInfo);
  } else
    _this->hidden->fFullScreenMode = 0;

  // Set buffer pointer and other remaining stuffs
  current->pitch = _this->hidden->ulScanLineBytes;
  current->pixels = _this->hidden->pBuffer;

  // Now destroy the message queue, if we've created it!
  if (ERRORIDERROR(hmqerror)==0)
  {
#ifdef DEBUG_BUILD
    printf("Destroying out message queue\n"); fflush(stdout);
#endif
    WinDestroyMsgQueue(hmq);
  }

#ifdef DEBUG_BUILD
  printf("SetVideoMode done\n"); fflush(stdout);
#endif
  /* We're done */
  return(current);
}


/* List the available video modes for the given pixel format, sorted
 from largest to smallest.
 */
static SDL_Rect **os2video_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
  PGDDMODEINFO pModeInfo;
  int iModeNum;

  printf("ListModes of %d Bpp\n", format->BitsPerPixel);

  // Destroy result of previous call, if there is any
  if (_this->hidden->pListModesResult)
  {
    free(_this->hidden->pListModesResult); _this->hidden->pListModesResult = NULL;
  }

#ifdef ALLOW_EVERY_MODE_IN_WINDOWED
  // Check if they need fullscreen or non-fullscreen video modes!
  if ((flags & SDL_FULLSCREEN) == 0)

  {
    // For windowed mode we support every resolution!
    return (SDL_Rect **)-1;
  } else
#endif
  {
    // For fullscreen mode we don't support every resolution!
#ifdef DEBUG_BUILD
//    printf("[os2video_ListModes] : Available fullscreen modes:\n");
#endif
    // Now create a new list
    pModeInfo = _this->hidden->ModeInfo;
    for (iModeNum=0; iModeNum<_this->hidden->NumVideoModes; iModeNum++)
    {
#ifdef DEBUG_BUILD
//      PrintModeInfo(pModeInfo);
#endif
      if (pModeInfo->ulBpp == format->BitsPerPixel)
      {
        SDL_Rect *pRect = (SDL_Rect *) malloc(sizeof(SDL_Rect));
        if (pRect)
        {
          // Fill description
          pRect->x = 0;
          pRect->y = 0;
          pRect->w = pModeInfo->ulHorizResolution;
          pRect->h = pModeInfo->ulVertResolution;
#ifdef DEBUG_BUILD
//          printf("!!! Seems to be good!\n");
//	  printf("F: %dx%d\n", pRect->w, pRect->h);
#endif
          // And insert into list of pRects
          if (!(_this->hidden->pListModesResult))
          {
#ifdef DEBUG_BUILD
//            printf("!!! Inserting to beginning\n");
#endif

            // We're the first one to be inserted!
            _this->hidden->pListModesResult = (SDL_Rect**) malloc(2*sizeof(SDL_Rect*));
            if (_this->hidden->pListModesResult)
            {
              _this->hidden->pListModesResult[0] = pRect;
	      _this->hidden->pListModesResult[1] = NULL;
            } else
            {
              free(pRect);
            }
          } else
          {
            // We're not the first ones, so find the place where we
            // have to insert ourselves
            SDL_Rect **pNewList;
            int iPlace, iNumOfSlots, i;

#ifdef DEBUG_BUILD
//            printf("!!! Searching where to insert\n");
#endif

            iPlace = -1; iNumOfSlots = 1; // Count the last NULL too!
            for (i=0; _this->hidden->pListModesResult[i]; i++)
            {
              iNumOfSlots++;
              if (iPlace==-1)
              {
                if ((_this->hidden->pListModesResult[i]->w*_this->hidden->pListModesResult[i]->h)<
                    (pRect->w*pRect->h))
                {
                  iPlace = i;
                }
              }
            }
            if (iPlace==-1) iPlace = iNumOfSlots-1;

#ifdef DEBUG_BUILD
//            printf("!!! From %d slots, it will be at %d\n", iNumOfSlots, iPlace);
#endif

            pNewList = (SDL_Rect**) realloc(_this->hidden->pListModesResult, (iNumOfSlots+1)*sizeof(SDL_Rect*));
            if (pNewList)
            {
              for (i=iNumOfSlots;i>iPlace;i--)
                pNewList[i] = pNewList[i-1];
              pNewList[iPlace] = pRect;
              _this->hidden->pListModesResult = pNewList;
            } else
            {
              free(pRect);
            }
          }

        }
      }
      pModeInfo++;
    }
  }
#ifdef DEBUG_BUILD
//  printf("Returning list\n");
#endif
  return _this->hidden->pListModesResult;
}

/* Initialize the native video subsystem, filling 'vformat' with the
 "best" display pixel format, returning 0 or -1 if there's an error.
 */
static int os2video_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
  int rc, iBPP;
  INITPROCOUT ipo;
  long ModeOperation;

  // Load VMAN.DLL
  rc = DosLoadModule(NULL, 0, "VMAN", &(_this->hidden->hmodVMAN));
  if (rc!=NO_ERROR) 
  {
    SDL_SetError("Could not load VMAN.DLL");
    return -1;
  }

  // Query entry point
  rc = DosQueryProcAddr(_this->hidden->hmodVMAN, 0, "VMIEntry", (PFN *)(&(_this->hidden->pfnVMIEntry)));
  if (rc!=NO_ERROR)
  {
    // Bad VMAN.DLL?
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : Could not query VMIEntry address! rc = 0x%x\n", rc);
    fflush(stdout);
#endif
    SDL_SetError("Could not query VMIEntry address in VMAN.DLL");
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  // Send an InitProc command to VMAN, informing it
  // about a new incoming process to use its services
  rc = _this->hidden->pfnVMIEntry(0, VMI_CMD_INITPROC, NULL, (PVOID) &ipo);
  if (rc!=NO_ERROR)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : VMI_CMD_INITPROC rc = 0x%x\n", rc);
    fflush(stdout);
#endif
    SDL_SetError("Error at VMI_CMD_INITPROC");
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }
  _this->hidden->fInFullScreenNow = FALSE;

  // Query the current video mode, and
  // store it in DesktopModeInfo
  rc = _this->hidden->pfnVMIEntry(0, VMI_CMD_QUERYCURRENTMODE, NULL, &(_this->hidden->DesktopModeInfo));
  if (rc!=NO_ERROR)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : Could not query DesktopModeInfo! rc = 0x%x\n", rc);
    fflush(stdout);
#endif
    SDL_SetError("Error at VMI_CMD_QUERYCURRENTMODE");
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  // Query the number of available video
  // modes to NumVideoModes, allocate memory
  // for information of all the video modes,
  // and query video mode information into
  // ModeInfo.
  ModeOperation = QUERYMODE_NUM_MODES;
  rc = _this->hidden->pfnVMIEntry(0, VMI_CMD_QUERYMODES, &ModeOperation, &(_this->hidden->NumVideoModes));
  if (rc!=NO_ERROR)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : VMI_CMD_QUERYMODES rc = 0x%x\n", rc);
    fflush(stdout);
#endif
    SDL_SetError("Error at VMI_CMD_QUERYMODES for QUERYMODE_NUM_MODES");
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  _this->hidden->ModeInfo = (PGDDMODEINFO) malloc(_this->hidden->NumVideoModes * sizeof(GDDMODEINFO));
  if (!(_this->hidden->ModeInfo))
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : Could not allocate memory for ModeInfo list!\n");
    fflush(stdout);
#endif
    SDL_OutOfMemory();
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  ModeOperation = QUERYMODE_MODE_DATA;
  rc = _this->hidden->pfnVMIEntry(0, VMI_CMD_QUERYMODES, &ModeOperation, _this->hidden->ModeInfo);
  if (rc!=NO_ERROR)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : Could not query ModeInfo list! rc = 0x%x\n", rc);
    fflush(stdout);
#endif
    SDL_SetError("Error at VMI_CMD_QUERYMODES for QUERYMODE_NUM_MODES");
    free(_this->hidden->ModeInfo); _this->hidden->ModeInfo = NULL;
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  // Now report the best pixel format. For this,
  // we'll use the current desktop format.

  /* Determine the screen depth */
  vformat->BitsPerPixel = _this->hidden->DesktopModeInfo.ulBpp;
  vformat->BytesPerPixel = (vformat->BitsPerPixel+7)/8;

#ifdef DEBUG_BUILD
  printf("[os2video_VideoInit] : Desktop mode :\n");
  fflush(stdout);
  PrintModeInfo(&(_this->hidden->DesktopModeInfo));
#endif

  GetBitMask(_this->hidden->DesktopModeInfo.fccColorEncoding,
	     &(vformat->Rmask),
	     &(vformat->Gmask),
             &(vformat->Bmask), &iBPP);

  // Fill in some window manager capabilities
  _this->info.wm_available = 1;

  // Initialize some internal variables
  _this->hidden->pListModesResult = NULL;

  // Now create our window with a default size
  // For this, we need a message processing thread.
  // We'll create a new thread for this, which will do everything
  // what is related to PM

  _this->hidden->iPMThreadStatus = 0;
  _this->hidden->tidPMThread = _beginthread(PMThreadFunc, NULL, 65536, (void *) _this);
  if (_this->hidden->tidPMThread == -1)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : Could not create PM thread!\n");
#endif
    SDL_SetError("Could not create PM thread");
    free(_this->hidden->ModeInfo); _this->hidden->ModeInfo = NULL;
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }
  // Burst the priority of PM Thread!
  DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 20, _this->hidden->tidPMThread);
  // Wait for the PM thread to initialize!
  while (_this->hidden->iPMThreadStatus==0)
    DosSleep(32);
  // If the PM thread could not set up everything, then
  // report an error!
  if (_this->hidden->iPMThreadStatus!=1)
  {
#ifdef DEBUG_BUILD
    printf("[os2video_VideoInit] : PMThread reported an error : %d\n", _this->hidden->iPMThreadStatus);
#endif
    SDL_SetError("Error initializing PM thread");
    free(_this->hidden->ModeInfo); _this->hidden->ModeInfo = NULL;
    _this->hidden->pfnVMIEntry(0, VMI_CMD_TERMPROC, NULL, NULL);
    DosFreeModule(_this->hidden->hmodVMAN);
    return -1;
  }

  return 0;
}


static void os2video_DeleteDevice(_THIS)
{
#ifdef DEBUG_BUILD
  printf("[os2video_DeleteDevice]\n"); fflush(stdout);
#endif
  // Free used memory
  free(_this->hidden);
  free(_this);
}

static int os2video_Available(void)
{
  int rc;
  HMODULE hmodVMAN;

  // We need Video Manager (VMAN) for querying list of available
  // video modes, and to be able to switch to fullscreen.
  // So, we'll check for the availability of VMAN!
  // We also need DIVE, but our code will be linked against it, so
  // if DIVE.DLL is not available, we won't even start. :)

#ifdef DEBUG_BUILD
  printf("[os2video_Available] : Trying to load VMAN.DLL\n");
#endif

  rc = DosLoadModule(NULL, 0, "VMAN", &hmodVMAN);            // Load VMAN.DLL
  if (rc!=NO_ERROR) 
  {
    // No VMAN.DLL... Maybe no GRADD driver installed???
#ifdef DEBUG_BUILD
    printf("[os2video_Available] : \n");
    printf("  Could not load VMAN.DLL! rc = 0x%x\n", rc);
    printf("  Are you sure you have GRADD video driver?\n");
#endif
    SDL_SetError("Could not load VMAN.DLL");
    return 0;
  }
  DosFreeModule(hmodVMAN);

#ifdef DEBUG_BUILD
  printf("[os2video_Available] : OK, os2video is available!\n");
#endif

  return 1;
}

static SDL_VideoDevice *os2video_CreateDevice(int devindex)
{
  SDL_VideoDevice *device;

  /* Initialize all variables that we clean on shutdown */
  device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
  if ( device )
  {
    memset(device, 0, (sizeof *device));
    device->hidden = (struct SDL_PrivateVideoData *) malloc((sizeof *device->hidden));
  }
  if ( (device == NULL) || (device->hidden == NULL) )
  {
    SDL_OutOfMemory();
    if ( device )
      free(device);
    return NULL;
  }
  memset(device->hidden, 0, (sizeof *device->hidden));

  /* Set the function pointers */

  /* Initialization/Query functions */
  device->VideoInit = os2video_VideoInit;
  device->ListModes = os2video_ListModes;
  device->SetVideoMode = os2video_SetVideoMode;
  device->ToggleFullScreen = os2video_ToggleFullScreen;
  device->UpdateMouse = os2video_UpdateMouse;
  device->CreateYUVOverlay = NULL;
  device->SetColors = os2video_SetColors;
  device->UpdateRects = os2video_UpdateRects;
  device->VideoQuit = os2video_VideoQuit;
  /* Hardware acceleration functions */
  device->AllocHWSurface = os2video_AllocHWSurface;
  device->CheckHWBlit = NULL;
  device->FillHWRect = NULL;
  device->SetHWColorKey = NULL;
  device->SetHWAlpha = NULL;
  device->LockHWSurface = os2video_LockHWSurface;
  device->UnlockHWSurface = os2video_UnlockHWSurface;
  device->FlipHWSurface = NULL;
  device->FreeHWSurface = os2video_FreeHWSurface;
  /* Window manager functions */
  device->SetCaption = os2video_SetCaption;
  device->SetIcon = os2video_SetIcon;
  device->IconifyWindow = os2video_IconifyWindow;
  device->GrabInput = NULL;
  device->GetWMInfo = NULL;
  /* Cursor manager functions */
  device->FreeWMCursor = os2video_FreeWMCursor;
  device->CreateWMCursor = os2video_CreateWMCursor;
  device->ShowWMCursor = os2video_ShowWMCursor;
  device->WarpWMCursor = os2video_WarpWMCursor;
  device->MoveWMCursor = os2video_MoveWMCursor;
  device->CheckMouseMode = os2video_CheckMouseMode;
  /* Event manager functions */
  device->InitOSKeymap = os2video_InitOSKeymap;
  device->PumpEvents = os2video_PumpEvents;
  /* The function used to dispose of this structure */
  device->free = os2video_DeleteDevice;

  return device;
}

VideoBootStrap OS2Video_bootstrap = {
  	"os2video", "OS/2 Digital Interface Video Extensions (DIVE)",
	os2video_Available, os2video_CreateDevice
};

